/*
 * main.c - Milena TTS Web Service
 * Copyright (C) Bohdan R. Rau 2008 <ethanak@polip.com>
 * 
 * Milena is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * Milena is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with Milena.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <getopt.h>

#include "milena_ws.h"
 
int conf_daemon=0;
int conf_port=20433;
int conf_autostop=1;
int conf_udecode=0;
char *conf_audio_driver;
char *pitch="0.8";
char *tempo="0.9";
struct say_message {
	struct say_message *next;
	char txt[1];
} *say_queue;

static int speak_pid;
static void say_direct(char *str)
{
	speak_pid=fork();
	if (speak_pid < 0) {
		perror("fork");
		exit(1);
	}
	if (!speak_pid) {
		say(str,ao_callback);
		exit(0);
	}
}

static int get_unichar(char *str,char **ptr)
{
	wchar_t wc;int n;
	wc=*str++ & 255;
	if ((wc & 0xe0)==0xc0) {
		wc &=0x1f;
		n=1;
	}
	else if ((wc & 0xf0)==0xe0) {
		wc &=0x0f;
		n=2;
	}
	else if ((wc & 0xf8)==0xf0) {
		wc &=0x07;
		n=3;
	}
	else if ((wc & 0xfc)==0xf8) {
		wc &=0x03;
		n=4;
	}
	else if ((wc & 0xfe)==0xfc) {
		wc &=0x01;
		n=5;
	}
	else {
		if (ptr) *ptr=str;
		return wc;
	}
	
	while (n--) {
		if ((*str & 0xc0) != 0x80) {
			wc='?';
			break;
		}
		wc=(wc << 6) | ((*str++) & 0x3f);
	}
	if (ptr) *ptr=str;
	return wc;
}

static void speak(char *str)
{
	struct say_message *sm,**ssm;
	int len=to_iso2(str,NULL);
	if (len<2) return;
	sm=malloc(len+sizeof(*sm));
	if (!sm) return;
	to_iso2(str,sm->txt);
	for (ssm=&say_queue;*ssm;ssm=&(*ssm)->next);
	*ssm=sm;
	sm->next=NULL;
}

static void get_queue(void)
{
	struct say_message *sm;
	if (sm=say_queue) {
		say_queue=sm->next;
		say_direct(sm->txt);
		free(sm);
	}
}

static void empty_queue(void)
{
	struct say_message *sm;
	while (sm=say_queue) {
		say_queue=sm->next;
		free(sm);
	}
}

void stop_speak(void)
{
	if (speak_pid) kill(speak_pid,9);
	empty_queue();
}

static void udecode(char *c)
{
	char *d,ss[4];
	for (d=c;*c;) {
		if (*c != '%') {
			if (*c=='+') *d++=' ';
			else *d++=*c;
			c++;
			continue;
		}
		c++;
		if (!*c || !c[1] || !isxdigit(*c) || !isxdigit(c[1])) {
			*d++='%';
			continue;
		}
		ss[0]=*c++;
		ss[1]=*c++;
		ss[2]=0;
		*d++=strtol(ss,0,16);
	}
	*d=0;
}

void read_http(int fd)
{
	FILE *f;
	int conlen;
	char *cs,*txt;
	char linebuf[2048];
	f=fdopen(fd,"r+");
	if (!f) return;
	if (!fgets(linebuf,2048,f)) {
		fclose(f);
		return;
	}
	conlen=0;
	if (strncasecmp(linebuf,"POST",4)) {
		while(fgets(linebuf,2048,f)) if (linebuf[0]=='\r' || linebuf[0]=='\n') break;
		fprintf(f,"HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<html><body><p>Milena Server</p></body></html>");
		fclose(f);
		return;
	}
	while(fgets(linebuf,2048,f)) {
		if (linebuf[0]=='\r' || linebuf[0]=='\n') break;
		if (!strncasecmp(linebuf,"Content-Length:",15)) conlen=strtol(linebuf+15,0,10);
	}
	txt="";
	if (!conlen) cs=NULL;
	else {
		cs=malloc(conlen+1);
		fread(cs,conlen,1,f);
		cs[conlen]=0;
		if (!strncasecmp(cs,"isSpeaking",10)) {
			txt=speak_pid?"True":"False";
		}
	}
	fprintf(f,"HTTP/1.0 200 OK\r\nContent-Type: text/plain\r\n\r\n%s",txt);	
	fclose(f);
	if (cs) {
		if (!strncasecmp(cs,"speak:",6)) {
			if (conf_autostop) stop_speak();
			if (conf_udecode) udecode(cs+6);
			speak(cs+6);
		}
		else if (!strcasecmp(cs,"stop")) stop_speak();
		free(cs);
	}
	return;
}

void help(char *arg)
{
	fprintf(stderr,"Sposob uzycia: %s [opcje]\nOpcje:\n",arg);
	fprintf(stderr," -h\thelp\n");
	fprintf(stderr," -T\ttest ustawienia mowy\n");
	fprintf(stderr," -p <port>\tport serwera (dom. 20433)\n");
	fprintf(stderr," -e\temuluj EmacsSpeak na porcie 2222\n");
	fprintf(stderr," -d <driver>\tdriver ao\n");
	fprintf(stderr," -a <opcja>=<wartosc>\topcja ao\n");
	fprintf(stderr," -D\tdaemon (w tle)\n");
	fprintf(stderr," -f <freq>\n -t <tempo>\tparametry dla mbroli (patrz mbrola -h\n");
	fprintf(stderr,"Drivery ao:\n");
	ao_initialize();
	ao_list_drivers();
	exit(0);
}


int main(int argc,char *argv[])
{
	int gniazdko;
	struct sockaddr_in sina;
	int reuse=-1;
	ignore_oor=1;
	int test_only=0;

	for(;;) {
		int c=getopt(argc,argv,"hp:d:a:Df:t:eT");
		if (c<0) break;
		if (c=='p') {
			conf_port=strtol(optarg,NULL,10);
			continue;
		}
		if (c=='d') {
			conf_audio_driver=optarg;
			continue;
		}
		if (c=='a') {
			ao_add_option(optarg);
			continue;
		}
		if (c=='D') {
			conf_daemon=1;
			continue;
		}
		if (c=='f') {
			pitch=optarg;
			continue;
		}
		if (c=='t') {
			tempo=optarg;
			continue;
		}
		if (c=='e') {
			conf_port=2222;
			conf_autostop=0;
			continue;
		}
		if (c=='T') {
			test_only=1;
			continue;
		}
		help(argv[0]);
	}
	if (optind != argc) help(argv[0]);
	if (test_only) {
		start_milena();
		start_audio();
		say_direct("Witam. Nazywam się Milena. Jestem systemem przetwarzania tekstu na mowę");
		if (speak_pid) waitpid(speak_pid,NULL,0);
		exit(0);
	}
	
	if (conf_daemon) {
		close(0);
		close(1);
		close(2);
		if (fork()) exit(0);
	}
	
	gniazdko=socket(PF_INET,SOCK_STREAM,0);
	if (gniazdko<0) {
		perror("socket");
		exit(0);
	}
	bzero(&sina,sizeof(sina));
	sina.sin_family = AF_INET;
	sina.sin_port = htons(conf_port);
	fcntl(gniazdko,F_SETFL,O_NONBLOCK);
	setsockopt(gniazdko,SOL_SOCKET,SO_REUSEADDR,(void *)&reuse,sizeof(reuse));
	if (bind(gniazdko, (struct sockaddr *)&sina,sizeof(sina))) {
		perror("bind");
		close(gniazdko);
		exit(0);
	}
	if (listen(gniazdko,128)) {
		perror("listen");
		close(gniazdko);
		exit(0);
	}
	start_milena();
	start_audio();
	for(;;) {
		fd_set rdfs;
		int retval;
		struct timeval tv;
		int fd;
		if (speak_pid) {
			retval=waitpid(speak_pid,NULL,WNOHANG);
			if (retval == speak_pid) {
				speak_pid=0;
			}
		}
		if (!speak_pid) get_queue();
		FD_ZERO(&rdfs);
		FD_SET(gniazdko,&rdfs);
		tv.tv_sec = 0;
		tv.tv_usec=100000;
		
		retval=select(gniazdko+1,&rdfs,0,0,&tv);
		fd=accept(gniazdko,0,0);
		if (fd<0) continue;
		read_http(fd);
		listen(gniazdko,128);
	}
	return 0;
}
